#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/event.h>

#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <err.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>
#include <time.h>

#define MAX_USERS    100   /* default */
#define MSG_SIZE     1500  /* default */
#define DEF_QUEUE    8128  /* default */

int failed = 0, all = 0;

int queue = 0;     /* size of client socket queue */
int msg_size = 0;  /* size of message to read from server and to write to clients */
int max_users = 0;

void 
writeLogs(int priority, const char *message) {
       openlog("multistream",LOG_PID | LOG_NDELAY, LOG_DAEMON);
       syslog(priority, message);
       closelog(); 
}

void 
sig_handler(int signo) {
    if (signo == SIGUSR1) printf("\nStatistics:
\ttotal blocks:\t%d\t100 percent
\twrited blocks:\t%d\t%5.2f percent
\tfailed blocks:\t%d\t%5.2f percent
block size:\t %d bytes\n\n",
all, all - failed, all ? (float)(all - failed) / all * 100 : all,
failed, all ? (float)failed / all * 100 : all, msg_size);
    return;
}

int
main(int ac, char **av) {
       
	int serverSocket, clientSocket, userSocket;
        struct sockaddr_in serverAddr, clientAddr, userAddr;
        int clientPort, serverPort, len, i, j, userCounter = 0, debug = 0;
        char *p, *msg = "Too many users\n";
        struct in_addr serverIP;
        struct hostent *he;
	int *user_fd /* users sockets */;
        
	int kq, event;
	struct kevent ke;

	char *buf; /* buffer for read from server and write to clients */
	char readBuf[10000];
        char logMessage[256];
	char *getVideo="GET /axis-cgi/mjpg/video.cgi?camera=1&resolution=352x288&compression=60 HTTP/1.1\n\n";
	char *getAudio="GET /axis-cgi/audio/receive.cgi HTTP/1.1\n\n";
	char *toVideoClient="HTTP/1.0 200 OK
Connection: Close
Server: Camd
Content-Type: multipart/x-mixed-replace; boundary=--myboundary

--myboundary

";
	char *toAudioClient="HTTP/1.0 200 OK
Content-type: multipart/x-mixed-replace; boundary=myboundary


";
	FILE *pid = NULL;

        time_t time1 = 0, time2 = 0;

	int keep_alive = 1;
	int reuse_addr = 1;
	int sound = 0;

        signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
        
	signal(SIGUSR1, &sig_handler);	

        while ( (i = getopt(ac, av, "dsP:u:q:m:")) > 0) {
                switch (i) {
                case 'd':
                       debug = 1;
                break;
		case 's':
		       sound = 1;
		break;
		case 'P':
        		if ((pid = fopen(optarg, "w")) == NULL) 
            			err(1, "Can't create PID file %s", optarg);
        	break;
        	case 'u':
        	        max_users = atoi(optarg);
        	break;
        	case 'q':
        	        queue = atoi(optarg);
        	break;
        	case 'm':
        	        msg_size = atoi(optarg);
        	break;
		default:
			err(1, "Unknown parameter");
                }   
        }

        ac -=optind;
        av +=optind;

        if (ac < 2) {
                printf("Usage: multistream [-d] [-s] [-P pidfile] [-u max_users] [-q queue] [-m msg_size] client_port host:port
\t-d - not daemonize, debug mode,
\t-s - get audio stream (default: video stream)
\t-u - max users to connect to multistream (default: %d)
\t-q - size of user sockets output queue (default: %d)
\t-m - size of message to read from server and to write to user sockets (default: %d)\n", MAX_USERS, DEF_QUEUE, MSG_SIZE);
                exit(0);
        }
        
        if (!max_users) max_users = MAX_USERS;
        if ((user_fd = (int *)malloc(max_users * sizeof(int))) < 0)
                err(1, "malloc");
        
        if (!queue) queue = DEF_QUEUE;
        
        if (!msg_size) msg_size = MSG_SIZE;
        if ((buf = (char *)malloc(msg_size * sizeof(char))) < 0)
                err(1, "malloc");
        
        clientPort = atoi(av[0]);
               
        if ((p = strchr(av[1], ':')) != NULL) {
                *(p++) = '\0';
                serverPort = atoi(p);       
        } else {
                printf("Bad server port\n");
                exit(0);
        }
        
        if (!(he = gethostbyname(av[1]))) {
                printf("Unknown server name\n");
                exit(0);
        }   
     
        serverIP = *(struct in_addr *)he->h_addr_list[0];
        
        /* connect to server */
        if ((serverSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
		err(1, "serverSocket");
        memset(&serverAddr, 0, sizeof(struct sockaddr_in));
	serverAddr.sin_family = AF_INET;
        serverAddr.sin_port = htons(serverPort);
        serverAddr.sin_addr = serverIP;

	if (setsockopt(serverSocket, SOL_SOCKET, SO_KEEPALIVE, (char *) &keep_alive, sizeof(SO_KEEPALIVE)) < 0)
                err(1, "setsockopt");
	
	if (connect(serverSocket, (struct sockaddr *)&serverAddr, sizeof serverAddr) < 0) 
		err(1, "connect");
	
	if (!sound) i = write(serverSocket, getVideo, strlen(getVideo));
	else i = write(serverSocket, getAudio, strlen(getAudio));
	
	if (i <= 0) 
	       err(1, "write");
	       
	/* listening socket */
        if ((clientSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		err(1, "clientSocket");
	memset(&clientAddr, 0, sizeof(struct sockaddr_in));
	clientAddr.sin_family = AF_INET;
        clientAddr.sin_port = htons(clientPort);
        clientAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if (setsockopt(clientSocket, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse_addr, sizeof(reuse_addr)) < 0) 
		err(1, "setsockopt");

	if (bind(clientSocket, (struct sockaddr *)&clientAddr, sizeof clientAddr) < 0)
		err(1, "bind");
	
	if (listen(clientSocket, 1) < 0)	
		err(1, "listen");

        /* daemonize */
	if  (!debug && daemon(debug, debug) < 0) {
                err(1, "can't daemonize");
        }

	/* kqueue */
	if ((kq = kqueue()) < 0)
		err(1, "kqueue");	
	
	/* fill out the kevent struct */
	memset(&ke, 0, sizeof(struct kevent));
	EV_SET(&ke, serverSocket, EVFILT_READ, EV_ADD, 0, 5, NULL);

	/* set the event */
	if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0)
		err(1, "kevent: serverSocket");
	
	/* fill out the kevent struct */
	memset(&ke, 0, sizeof(struct kevent));
	EV_SET(&ke, clientSocket, EVFILT_READ, EV_ADD, 0, 5, NULL);

	/* set the event */
	if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0)
		err(1, "kevent: clientSocket");

	memset(user_fd, 0, sizeof(int) * MAX_USERS);
	
	if (pid) {
        	fprintf(pid, "%d", getpid());
        	fclose(pid);
    	}

	if (debug) 
	           time1 = time(NULL);
/*******************************************************/	
	if (debug) printf("Now ready to accept users!
******* INFO ******* 
stream:\t\t%s
server:\t\t%s:%d
client port:\t%d
max users:\t%d
output queue\t%d
msg size:\t%d
debug mode:\t%s
*******************\n", sound ? "audio" : "video", inet_ntoa(serverIP), serverPort, clientPort, max_users, queue, msg_size, debug ? "yes" : "no");
/*******************************************************/
	
	while(1) {
		
		if (debug) {
		          time2 = time(NULL);  
		          if (abs(time2 - time1) >= 10) {
		                  time1 = time2;
		                  printf("\nStatistics:
\ttotal blocks:\t%d\t100 percent
\twrited blocks:\t%d\t%5.2f percent
\tfailed blocks:\t%d\t%5.2f percent
block size:\t %d bytes\n\n",
all, all - failed, all ? (float)(all - failed) / all * 100 : all,
failed, all ? (float)failed / all * 100 : all, msg_size);
for(i = 0; i < userCounter; i++) 	
	printf("user_fd[%d] = %d\n", i, user_fd[i]);
		           }
		}
		memset(&ke, 0, sizeof(ke));
                  
		/* receive an event, a blocking call as timeout is NULL */
		if ((event = kevent(kq, NULL, 0, &ke, 1, NULL)) < 0) {
			if (errno == EINTR) continue;
			else {
			     snprintf(logMessage, 256, "kevent: %s", strerror(errno));
                             writeLogs(LOG_EMERG, logMessage);
			     err(1, "kevent");
		        }		
		} 
                
                if (event == 0) continue;
		
		/* got a message from clientSocket */
		if (ke.ident == clientSocket) {
			len = (socklen_t)sizeof(userAddr);
			userSocket = accept(clientSocket, (struct sockaddr *)&userAddr, &len);
			if (userSocket == -1) {
			        snprintf(logMessage, 256, "accept: %s", strerror(errno));
                                writeLogs(LOG_ALERT, logMessage); 
				continue;
			}
			
			if (setsockopt(userSocket, SOL_SOCKET, SO_SNDBUF, (char *)&queue, sizeof(queue)) < 0) {
				snprintf(logMessage, 256, "setsockopt: %s", strerror(errno));
                                writeLogs(LOG_ALERT, logMessage); 
				continue;
			}			
			if (userCounter == max_users) {
				write(userSocket, msg, strlen(msg));
				if (debug) printf("User limit reached\n");
				snprintf(logMessage, 256, "User limit reached");
                                writeLogs(LOG_NOTICE, logMessage);
				close(userSocket);
				continue;
			}
		
			/* send user server reply */
			if (write(userSocket, sound ? toAudioClient : toVideoClient, strlen(sound ? toAudioClient : toVideoClient)) <= 0) {
			            close(userSocket);
			            continue;
			}
			
			if (debug) printf("New user connected:\t%s:%d\n", inet_ntoa(*(struct in_addr *)&userAddr.sin_addr.s_addr),
			                ntohs(userAddr.sin_port));
                        snprintf(logMessage, 256, "New user connected: %s:%d", inet_ntoa(*(struct in_addr *)&userAddr.sin_addr.s_addr), 
                                  ntohs(userAddr.sin_port));
                        writeLogs(LOG_NOTICE, logMessage);
			
			EV_SET(&ke, userSocket, EVFILT_READ, EV_ADD, 0, 0, NULL);
			if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0) {
			        snprintf(logMessage, 256, "kevent: add user: %s", strerror(errno));
                                writeLogs(LOG_EMERG, logMessage);
				continue;
		        }
		        
		        user_fd[userCounter++] = userSocket;
		/* got a message from server */
		} else if (ke.ident == serverSocket) {
			memset(buf, 0, sizeof(buf));

			i = read(serverSocket, buf, msg_size);
		 	//printf("From server: buf=\n%s\n",buf);		
		 			
			if (i == -1)
				continue;
			if (i == 0) { /* EOF from server */
				/* connection closed by server */
				if (debug) printf("Connection closed by server\n");
				snprintf(logMessage, 256, "Connection closed by server");
                                writeLogs(LOG_NOTICE, logMessage);
				close(serverSocket);
				exit(0);
			}
                        
                        for (j = 0; j < userCounter; j++) {
                                if (write(user_fd[j], buf, i) < 0) 
					failed++;
				all++;		
		        }
		} else 
		        /* got a message from users */
		        for (i = 0; i < userCounter; i++) 
		                if (ke.ident == user_fd[i]) {  
		                        if (read(user_fd[i], readBuf, sizeof(readBuf)) == 0) {
		                                /* delete user */
		                                EV_SET(&ke, user_fd[i], EVFILT_READ, EV_DELETE, 0, 0, NULL);
                			        if (kevent(kq, &ke, 1, 0, 0, NULL) < 0) {
                			                snprintf(logMessage, 256, "kevent: rm user from kq");
                                                        writeLogs(LOG_EMERG, logMessage); 
                				        err(1, "rm user from kq");
                				}
						close(user_fd[i]);
	                                        user_fd[i] = user_fd[--userCounter];
						if (debug) printf("User <%s:%d> logged off\n",
	                                                        inet_ntoa(*(struct in_addr *)&userAddr.sin_addr.s_addr),
			                                        ntohs(userAddr.sin_port));
						snprintf(logMessage, 256, "User <%s:%d> logged off",
						         inet_ntoa(*(struct in_addr *)&userAddr.sin_addr.s_addr),
			                                 ntohs(userAddr.sin_port));
                                                writeLogs(LOG_NOTICE, logMessage);
	                                 }
	                                 break;
	                        }
	} /* end of while */	
}
